import collections
import inspect
import re
from importlib import import_module
from pkgutil import iter_modules


def normalize_slash(tag, with_prefix_slash=True):
    tag = re.sub('[/]+', '/', tag)

    if with_prefix_slash and not tag.startswith('/'):
        tag = '/' + tag

    return tag


def walk_modules(module_path, skip_fail=True):

    mod = None
    try:
        mod = import_module(module_path)
        yield mod
    except Exception as e:
        if not skip_fail:
            raise e

    if mod and hasattr(mod, '__path__'):
        for _, subpath, _ in iter_modules(mod.__path__):
            fullpath = '.'.join((module_path, subpath))
            for m in walk_modules(fullpath, skip_fail):
                yield m


def expected_cls(module, cls, base_class, include_base_class=False):
    if inspect.isclass(cls) and \
            issubclass(cls, base_class) and \
            cls.__module__ == module.__name__ and \
            (include_base_class or
             all([cls != base for base in base_class])
             if isinstance(base_class, tuple)
             else cls != base_class):
        return True
    return False


def load_class(class_path, base_class, include_base_class=False, package=None):
    module_path, class_name = class_path.rsplit('.', 1)
    module = import_module(module_path, package=package)
    cls = getattr(module, class_name)
    if expected_cls(module, cls, base_class, include_base_class):
        return cls
    return None


def deep_update(d, u):
    for k, v in u.items():
        if isinstance(v, collections.Mapping):
            d[k] = deep_update(d.get(k, {}), v)
        else:
            d[k] = v
    return d



import os
import socket
from contextlib import closing, contextmanager
from shutil import copy2, copystat


@contextmanager
def cd(new_path):
    old_path = os.getcwd()
    yield os.chdir(new_path)
    os.chdir(old_path)


def copy_file(f, dest):
    path = os.path.dirname(__file__)
    f = os.path.join(path, f)
    copy2(f, dest)


def unused_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(('', 0))
        return s.getsockname()[1]



import inspect
import logging
import os
from functools import partial

from sanic import Sanic
from sanic.log import LOGGING_CONFIG_DEFAULTS

from os_sanic.application import ApplicationManager
from os_sanic.config import SANIC_ENV_PREFIX, create_sanic_config
from os_sanic.log import LOGGING_CONFIG_PATCH, logger
from os_sanic.utils import deep_update
from os_sanic.monkey_patch import patch


class Server(object):
    def __init__(self, app_manager, enable_workflow=True):
        self.app_manager = app_manager
        if enable_workflow:
            self.__register_workflow()

    def __register_workflow(self):
        [self.sanic.register_listener(partial(self.__call, method), event)
         for method, event in zip(('run', 'setup', 'cleanup'),
                                  ('before_server_start',
                                   'after_server_start',
                                   'after_server_stop'))]

    @property
    def sanic(self):
        return self.app_manager.sanic

    async def __call(self, method, app, loop):
        await getattr(self.app_manager, method)()

    def _run_args(self):
        argspec = inspect.getargspec(self.sanic.run)
        run_args = {}
        offset = len(argspec.args)-len(argspec.defaults)
        for idx in range(len(argspec.defaults)-1, -1, -1):
            default = argspec.defaults[idx]
            idx += offset
            name = argspec.args[idx]
            run_args[name] = self.sanic.config.get(name.upper(), default)

        return run_args

    def run(self):
        with patch():
            self.sanic.run(**self._run_args())

    @classmethod
    def create(cls, app='os-sanic', config=None, env_prefix=SANIC_ENV_PREFIX, log_config=None):

        assert isinstance(app, (str, Sanic))

        if isinstance(app, str):
            app = Sanic(app, load_env=env_prefix)

        conf = create_sanic_config(load_env=env_prefix)
        if config:
            conf.update(config)

        app.config.update(conf)

        if app.configure_logging:
            if not log_config:
                log_config = deep_update({}, LOGGING_CONFIG_DEFAULTS)
                deep_update(log_config, LOGGING_CONFIG_PATCH)
            logging.config.dictConfig(log_config)

        if app.config.get('DEBUG', False):
            if os.environ.get('SANIC_SERVER_RUNNING') != 'true':
                return cls(ApplicationManager.create(app), False)

            logger.setLevel('DEBUG')
            logger.debug('Debug mode')
        else:
            logger.setLevel(app.config.LOG_LEVEL)

        logger.debug(f'Config: {app.config}')

        return cls(ApplicationManager.create(app))



import inspect

from sanic import Sanic

from os_sanic.cmdline import execute


def run(app=None, command_packages=None, **kwargs):
    if app is None:
        app = Sanic('os-sanic')

    if command_packages is None:
        command_packages = ['os_sanic.commands.project', ]

    execute(app=app, command_packages=command_packages, **kwargs)


def main():
    command_packages = ['os_sanic.commands.global', ]
    run(command_packages=command_packages)



import pytest
from os_sanic.server import Server


@pytest.fixture()
def os_sanic_server():

    def _create(sanic, config, *apps):
        if not config:
            config = {}
        if 'INSTALLED_APPS' not in config:
            config['INSTALLED_APPS'] = []

        config['INSTALLED_APPS'].extend(apps)
        return Server.create(sanic, config)

    yield _create



import os

import click

import os_sanic
from os_sanic.commands import create_from_tpl, valid_name
from os_sanic.config import create_sanic_config


def app_creation_params(app_name, base_path=None):
    app_package = f'apps.{app_name}'

    base_tpl_dir = os.path.join(
        os_sanic.__path__[0], 'commands', 'template')

    app_tpl_dir = os.path.join(base_tpl_dir, 'app_template')

    b = base_path if base_path is not None else os.getcwd()
    app_dst_dir = os.path.join(b, 'apps', f'{app_name}')

    return app_package, app_tpl_dir, app_dst_dir


def create_app(ctx, app_name, app_package, app_tpl_dir, app_dst_dir):

    if os.path.exists(app_dst_dir):
        ctx.fail(f'App already existed, {app_dst_dir}')

    apps_tpl_dir = app_tpl_dir.replace('app_template', 'apps_template')
    apps_dst_dir = app_dst_dir[0:app_dst_dir.rfind(app_name)]

    if not os.path.exists(apps_dst_dir):
        create_from_tpl(apps_tpl_dir, apps_dst_dir, ignores=['*.pyc', ])

    config = create_sanic_config()
    config.extension_class = config.extension_name = app_name.capitalize()
    config.uri = '/'
    config.view_class = config.extension_class + 'View'
    create_from_tpl(app_tpl_dir, app_dst_dir,
                    ignores=['*.pyc', ], **config)


@click.command()
@click.argument('app-name', callback=valid_name)
@click.pass_context
def cli(ctx, app_name):
    '''Create new application.'''

    app_package, app_tpl_dir, app_dst_dir = app_creation_params(app_name)
    create_app(ctx, app_name, app_package, app_tpl_dir, app_dst_dir)

    click.echo(f'New os-sanic app: {app_name}\n')
    click.echo('Use app template:')
    click.echo(f'    {app_tpl_dir}\n')
    click.echo('Create app in:')
    click.echo(f'    {app_dst_dir}\n')
    click.echo('You should add app package into INSTALLED_APPS:')
    click.echo(f'    \'{app_package}\'')



import click
from click import Command

from os_sanic.utils import walk_modules


class CommandFinder(click.MultiCommand):

    def list_commands(self, ctx):
        ctx.ensure_object(dict)
        return list(self.__find_commnds(**ctx.obj).keys())

    def get_command(self, ctx, name):
        ctx.ensure_object(dict)
        commands = self.__find_commnds(**ctx.obj)
        return commands.get(name, None)

    def __find_commnds(self, **kwargs):
        command_packages = kwargs.get('command_packages', [])
        commands = {}
        for command_package in command_packages:
            for cmd_module in walk_modules(command_package):
                if hasattr(cmd_module, 'cli') and isinstance(cmd_module.cli, Command):
                    commands[cmd_module.__name__.split(
                        '.')[-1]] = cmd_module.cli

        return commands


def execute(**kwargs):

    @click.command(cls=CommandFinder, context_settings=dict(obj=kwargs))
    @click.version_option(message='%(prog)s %(version)s')
    @click.pass_context
    def cli(ctx):
        '''Command line tool for os-sanic.'''
        pass

    cli()



from sanic.views import HTTPMethodView
from sanic.response import json


class Case02View(HTTPMethodView):

    def get(self, request):
        ext = self.application.get_extension('case01.case01')
        return json({'view': ext.test()})



from os_sanic.utils import deep_update, normalize_slash


def test_normalize_prefix():
    data = [
        ('/', '/'),
        ('a/', '/a/'),
        ('//a', '/a'),
        ('/a//', '/a/'),
        ('', '/'),
        ('/a', '/a'),
        ('/a///c', '/a/c'),
    ]

    for raw, expect in data:
        assert normalize_slash(raw) == expect


def test_deep_update():
    a = {'a': {'b': 1}}
    b = {'a': {'b': 2}}

    deep_update(a, b)
    assert a['a']['b'] == 2



#!/usr/bin/env python
from sanic import Sanic
from os_sanic import run

app = Sanic(__name__)

if __name__ == "__main__":
    # =========== for test coverage ===============
    import os
    import sys
    sys.path.insert(0, os.getcwd())
    if os.getenv('COVERAGE_PROCESS_START'):
        import coverage
        coverage.process_startup()
    # =============================================

    run(app)



import ast
import os
from importlib import import_module
from logging import _nameToLevel
from shutil import copy2, copystat, ignore_patterns

import click
from jinja2 import Template


def valid_name(ctx, param, value):
    if value is None:
        return None
    def _module_exists(module_name):
        try:
            import_module(module_name)
            return True
        except ImportError:
            return False
    try:
        ast.parse(f'{value} = None')
    except:
        raise click.BadParameter(f'Not a valid name, {value}')

    if value.startswith('_'):
        raise click.BadParameter(f'Not a valid name, {value}')
    elif _module_exists(value):
        raise click.BadParameter(f'Module already existed, {value}')

    return value


def valid_log_level(ctx, param, value):
    v = value.upper()
    if v not in _nameToLevel:
        choices = ' '.join(_nameToLevel.keys())
        raise click.BadParameter(
            f'Invalid choice: {value}. (choose from {choices})')
    return v

def copy_tpl(src, dst, ignores=[]):
    if os.path.exists(dst):
        raise IOError(f'Path already existed {dst}')
    ignore = ignore_patterns(*ignores)
    names = os.listdir(src)
    ignored_names = ignore(src, names)

    if not os.path.exists(dst):
        os.makedirs(dst)

    for name in names:
        if name in ignored_names:
            continue

        src_name = os.path.join(src, name)
        dst_name = os.path.join(dst, name)
        if os.path.isdir(src_name):
            copy_tpl(src_name, dst_name, ignores)
        else:
            copy2(src_name, dst_name)
    copystat(src, dst)


def render_tpl(dst_dir, **kwargs):
    template_files = []
    for directory, _, filenames in os.walk(dst_dir):
        for name in filenames:
            template_file = os.path.join(directory,  name)
            if not template_file.endswith('-tpl'):
                continue
            template_files.append(template_file)
    for template_file in template_files:
        with open(template_file, 'rb') as fp:
            raw = fp.read().decode('utf8')
        dst_file = template_file[0:-len('-tpl')]
        Template(raw).stream(**kwargs).dump(dst_file)
        os.remove(template_file)


def create_from_tpl(tpl_dir, dst_dir, ignores=[], **kwargs):
    copy_tpl(tpl_dir, dst_dir, ignores)
    render_tpl(dst_dir, **kwargs)



import os
import signal
from contextlib import contextmanager
from time import sleep
from unittest import mock

from sanic import reloader_helpers
from sanic.reloader_helpers import (_iter_module_files, kill_process_children,
                                    restart_with_reloader)


@contextmanager
def patch():
    with mock.patch.object(reloader_helpers, 'watchdog', patch_reloader_helpers_watchdog):
        yield


def patch_reloader_helpers_watchdog(sleep_interval):
    mtimes = {}
    worker_process = []
    signal.signal(
        signal.SIGTERM, lambda *args: kill_program_completly(
            worker_process)
    )
    signal.signal(
        signal.SIGINT, lambda *args: kill_program_completly(
            worker_process)
    )

    worker_process.append(restart_with_reloader())
    while True:
        for filename in _iter_module_files():
            try:
                mtime = os.stat(filename).st_mtime
            except OSError:
                continue

            old_time = mtimes.get(filename)
            if old_time is None:
                mtimes[filename] = mtime
                continue
            elif mtime > old_time:
                wp = worker_process.pop()
                kill_process_children(wp.pid)
                wp.terminate()
                worker_process.append(restart_with_reloader())
                mtimes[filename] = mtime
                break

        sleep(sleep_interval)


def kill_program_completly(proc):
    if not proc:
        kill_process_children(os.getpid())
        os._exit(0)
    proc = proc.pop()
    kill_process_children(proc.pid)
    proc.terminate()
    os._exit(0)



import json

def test_001(os_sanic_server):
    server = os_sanic_server(
        'test', None, 'tests.apps.case01')
    _, response = server.sanic.test_client.get('/case01')
    assert response.status == 200
    assert json.loads(response.body) == {'view': 'Case01'}


def test_002(os_sanic_server):
    server = os_sanic_server(
        'test', None, 'tests.apps.case01', 'tests.apps.case02')

    _, response = server.sanic.test_client.get('/case02')
    assert response.status == 200
    assert json.loads(response.body) == {'view': 'Case01'}


def test_003(os_sanic_server):
    server = os_sanic_server(
        'test', None,  'tests.apps.case02')

    _, response = server.sanic.test_client.get('/case02/static/hello.txt')
    assert response.status == 200
    assert response.body == b'hello'



import json
import os
import uuid
from shutil import copytree

import pytest
import requests
from pytest_xprocess import getrootdir
from xprocess import ProcessStarter, XProcess

from tests.cmd_runner import call
from tests.utils import cd, copy_file, unused_port


@pytest.fixture()
def xproc(request):
    rootdir = getrootdir(request.config)
    rootdir = rootdir.join(str(uuid.uuid1())[0:8])
    yield XProcess(request.config, rootdir)
    if rootdir.exists():
        rootdir.remove()


@pytest.fixture
def cov_env():
    env = os.environ.copy()
    env['COVERAGE_PROCESS_START'] = os.path.abspath('.coveragerc')
    env['COVERAGE_FILE'] = os.path.abspath('.coverage')
    return env


@pytest.fixture
def new_project(tmpdir):
    def _create_project(proj_name, app_name=None):
        create_project(tmpdir, proj_name, app_name)
        proj_path = os.path.join(tmpdir.strpath, proj_name)
        with cd(proj_path):
            copy_file('manage_for_test.py', 'manage.py')
        return proj_path

    yield _create_project


def create_project(tmpdir, proj_name, app_name=None):
    f = 'cmd_runner.py'
    runf = os.path.join(tmpdir.strpath, f)
    copy_file(f, runf)
    env = os.environ.copy()
    env['COVERAGE_PROCESS_START'] = os.path.abspath('.coveragerc')
    env['COVERAGE_FILE'] = os.path.abspath('.coverage')

    cmd = f'startproject {proj_name}'
    if app_name:
        cmd += f' --with-app {app_name}'

    with cd(tmpdir):
        call(runf, cmd, env)


def test_no_args(tmpdir):
    stdout, _ = call()
    assert b'startproject' in stdout
    assert b'startapp' not in stdout


def test_startproject_001(new_project):
    proj_name = 'xxx'
    app_name = 'example'
    proj_path = new_project(proj_name, app_name)

    files = [
        'manage.py',
        'config.py',
        'apps',
        'apps/__init__.py',
        f'apps/{app_name}/__init__.py',
        f'apps/{app_name}/app.py',
        f'apps/{app_name}/view.py',
        f'apps/{app_name}/extension.py',
    ]
    print(proj_path)
    for f in files:
        assert os.path.exists(os.path.join(proj_path, f))


def test_start_app(new_project, cov_env):
    proj_path = new_project('xxx')
    print(proj_path)
    with cd(proj_path):
        app_name = 'yyy'
        call('manage.py', f'startapp {app_name}', cov_env)
        prefix = f'apps/{app_name}/'
        for ff in ('', 'app.py', 'extension.py', '__init__.py', 'view.py'):
            ff = os.path.join(prefix, ff)
            assert os.path.exists(ff)


def test_info(new_project, cov_env):
    proj_name = 'xxx'
    app_name = 'example'
    proj_path = new_project(proj_name, app_name)
    print(proj_path)
    expect = [
        f'"name": "{app_name}",',
        f'"package": "apps.{app_name}",',
        f'<class \'apps.{app_name}.extension.{app_name.capitalize()}\'>',
        f'<class \'apps.{app_name}.view.{app_name.capitalize()}View\'>',
    ]
    with cd(proj_path):
        stdout, _ = call('manage.py', 'info', cov_env)
        for exp in expect:
            assert exp.encode() in stdout


@pytest.fixture
def run_server(xproc, cov_env):
    name = 'sanic_server'

    def _run_server(proj_path, *run_args):

        class Starter(ProcessStarter):
            pattern = '.*Starting\\s+worker'
            env = cov_env
            args = [
                'python',
                'manage.py',
                *run_args,
            ]

            def __init__(self, control_dir, process):
                self.control_dir = control_dir
                self.process = process

        copytree(proj_path, xproc.rootdir.join(name))
        xproc.ensure(name, Starter)
        return xproc

    yield _run_server

    xproc.getinfo(name).terminate()


def test_run_001(new_project, run_server):
    proj_name = 'xxx'
    app_name = 'example'
    proj_path = new_project(proj_name, app_name)
    port = unused_port()
    run_server(proj_path, 'run', '--port', f'{port}')
    url = f'http://127.0.0.1:{port}/example'
    r = requests.get(url)
    assert r.status_code == 200
    d = json.loads(r.content)
    assert d == {'view': 'ExampleView'}


def test_run_002(new_project, run_server):
    proj_name = 'xxx'
    app_name = 'example'
    proj_path = new_project(proj_name, app_name)
    port = unused_port()
    xproc = run_server(proj_path, 'run', '--port', f'{port}', '-l', 'DEBUG')
    info = xproc.getinfo('sanic_server')
    log = info.logpath.open().read()
    expects = ['[App.example.Example] [INFO] run',
               '[App.example.Example] [INFO] setup']
    for exp in expects:
        assert exp in log



import json
import os
from collections import OrderedDict

import click
from os_config import Config

from os_sanic.commands.project.run import create_server


def server_info(server):
    click.echo('Server:')
    out = json.dumps(server.sanic.config, indent=4)
    click.echo(f'{out}\n')


def apps_info(server):
    am = server.app_manager
    apps_info = []
    for app in am.apps:
        info = app_info(app)
        apps_info.append(info)

    if apps_info:
        out = json.dumps(apps_info, indent=4)
        click.echo('Applications:')
        click.echo(out)


def app_info(app):
    info = OrderedDict()
    info['name'] = app.name
    info.update(Config.to_dict(app.app_cfg))

    extensions = app.extension_manager.extensions

    if extensions:
        exts_info = []
        for extension in extensions:
            ext_info = OrderedDict()
            ext_info['name'] = extension.name
            ext_info['extension_class'] = str(extension.__class__)
            ext_info.update(Config.to_dict(extension.config))
            exts_info.append(ext_info)
        if exts_info:
            info['extensions'] = exts_info

    views = app.view_manager.views
    url_prefix = app.view_manager.blueprint.url_prefix
    url_prefix = '' if not url_prefix else url_prefix
    if views:
        views_info = []
        for view in views:
            view_info = OrderedDict()
            view_info['uri'] = url_prefix+view.uri
            view_info['view_class'] = str(view.view_cls)
            view_info.update(Config.to_dict(view.config))
            views_info.append(view_info)
        if views_info:
            info['views'] = views_info

    statics = app.view_manager.statics
    if statics:
        statics_info = []
        for static in statics:
            static_info = OrderedDict()
            static_info['uri'] = static.uri
            static_info['file_or_directory'] = static.file_or_directory
            static_info.update(static.kwargs)
            statics_info.append(static_info)

        if statics_info:
            info['statics'] = statics_info

    return info


@click.command()
@click.option('-c', '--config-file', default='config.py',
              show_default=True, type=click.File(mode='r'),
              help='Config file')
@click.pass_context
def cli(ctx, config_file):
    '''Show config details.'''

    ctx.ensure_object(dict)

    server = create_server(ctx.obj['app'], config_file=config_file)
    server_info(server)
    apps_info(server)



import logging
import sys

from sanic.log import logger


def getLogger(name):
    lg = logging.getLogger(name)
    if not lg.hasHandlers():
        if logger.hasHandlers():
            lg.addHandler(logger.handlers[0])
            lg.setLevel(logger.level)
    return lg


LOGGING_CONFIG_PATCH = dict(

    formatters={
        "generic": {
            "format": "%(asctime)s [%(name)s] [%(levelname)s] %(message)s",
            "datefmt": "[%Y-%m-%d %H:%M:%S]",
        },
    },
    handlers={
        "console": {
            "stream": sys.stderr
        },
        "access_console": {
            "stream": sys.stderr
        },
    },

)

BASE_LOGGING = {
    **LOGGING_CONFIG_PATCH['formatters']['generic']
}



import os
from collections import OrderedDict

from os_config import Config
from sanic import Blueprint
from sanic.views import HTTPMethodView

from os_sanic.utils import load_class, normalize_slash


class View(object):
    def __init__(self, uri, view_cls, config):
        self.uri = uri
        self.view_cls = view_cls
        self.config = config

    @classmethod
    def create(cls, application, blueprint, view_cfg, user_config):

        uri, view_class = view_cfg.uri, view_cfg.view_class
        config = Config.create()
        Config.update(config, view_cfg)
        Config.update(config, user_config)
        Config.pop(config, 'uri')
        Config.pop(config, 'view_class')

        package = None
        if view_class.startswith('.'):
            package = application.app_cfg.package
        view_cls = load_class(
            view_class, HTTPMethodView, package=package)

        view_cls.application = application

        kwargs = {}
        if len(config) > 0:
            kwargs['config'] = config

        blueprint.add_route(view_cls.as_view(**kwargs), uri)

        return cls(uri, view_cls, config)


class ViewManager(object):

    def __init__(self, application, blueprint):
        self.application = application
        self.blueprint = blueprint
        self._views = OrderedDict()
        self.logger = application.get_logger(self.__class__.__name__)

    @property
    def views(self):
        return self._views.values()

    @property
    def statics(self):
        return self.blueprint.statics

    def load_static(self, static_cfg, user_config):
        try:
            s = Config.to_dict(static_cfg)
            s.update(Config.to_dict(user_config))
            uri = s.pop('uri')

            real_uri = self._real_uri(uri)

            self.blueprint.static(real_uri, os.path.abspath(os.path.join(
                self.application.runtime_path, s.pop('file_or_directory'))), **s)
            static = self.statics[-1]
            pth = self.blueprint.url_prefix + static.uri
            self.logger.debug(
                f'Load static, {pth} {self.statics[-1]}')
        except Exception as e:
            self.logger.error(f'Load static error, {e}')

    def _real_uri(self, uri):
        real_uri = normalize_slash(uri, with_prefix_slash=False)
        url_prefix = self.blueprint.url_prefix

        warn = True if uri != real_uri else False
        if not url_prefix.endswith('/') and not real_uri.startswith('/'):
            warn = True
        elif url_prefix.endswith('/') and real_uri.startswith('/'):
            warn = True
            real_uri = real_uri[1:]
        if warn:
            self.logger.warn(
                f'Maybe misspell with prefix: \'{url_prefix}\' uri: \'{uri}\'')
        return real_uri

    def _load_view(self, view_cfg, user_configs):
        if isinstance(view_cfg, tuple):
            view_cfg = Config.create(uri=view_cfg[0], view_class=view_cfg[1])

        uri, view_class = view_cfg.uri, view_cfg.view_class
        user_config = user_configs.get(uri, Config.create())

        real_uri = self._real_uri(uri)
        view_cfg.uri = real_uri

        pth = self.blueprint.url_prefix+real_uri
        if real_uri in self._views:
            self.logger.warn(
                f'Skip, view already existed, {pth} {view_class}')
            return

        view = View.create(self.application, self.blueprint,
                           view_cfg, user_config)
        self._views[real_uri] = view
        self.logger.debug(f'Load view, {pth} {view.view_cls}')

    def load_view(self, view_cfg, user_configs):
        try:
            self._load_view(view_cfg, user_configs)
        except Exception as e:
            self.logger.error(f'Load view error, {e}')

    @classmethod
    def create(cls, application):
        logger = application.get_logger(cls.__name__)

        _prefix = Config.get(application.app_cfg, 'prefix',
                             f'/{application.name}')
        prefix = normalize_slash(_prefix)

        if not prefix == _prefix:
            logger.warn(f'Normalize prefix from \'{_prefix}\' to \'{prefix}\'')

        blueprint = Blueprint(application.name, url_prefix=prefix)

        view_manager = cls(application, blueprint)

        user_configs = {}
        for v in Config.get(application.user_config, 'VIEWS', []):
            if not isinstance(v, Config):
                continue
            uri = Config.get(v, 'uri')
            if uri:
                Config.pop(v, 'uri')
                user_configs[uri] = v
        for view_cfg in Config.get(application.core_config, 'VIEWS', []):
            view_manager.load_view(view_cfg, user_configs)

        static_configs = {}
        for v in Config.get(application.user_config, 'STATICS', []):
            uri = Config.get(v, 'uri')
            if uri:
                Config.pop(v, 'uri')
                static_configs[uri] = v
        for static_cfg in Config.get(application.core_config, 'STATICS', []):
            view_manager.load_static(static_cfg, static_configs.get(
                static_cfg.uri, Config.create()))

        application.sanic.blueprint(blueprint)

        return view_manager



import os
import sys

import click

import os_sanic
from os_sanic.commands import create_from_tpl, valid_name
from os_sanic.commands.project.startapp import app_creation_params, create_app
from os_sanic.config import SANIC_ENV_PREFIX, create_sanic_config


@click.command()
@click.argument('project-name', callback=valid_name)
@click.option('--with-app', callback=valid_name, required=False, help='Create project with app.')
@click.pass_context
def cli(ctx, project_name, with_app):
    '''Create new project.'''

    base_tpl_dir = os.path.join(
        os_sanic.__path__[0], 'commands', 'template')

    proj_tpl_dir = os.path.join(base_tpl_dir, 'project_template')

    proj_dst_dir = os.path.join(os.getcwd(), project_name)

    if os.path.exists(proj_dst_dir):
        ctx.fail(f'Project already existed, {proj_dst_dir}')

    config = create_sanic_config(load_env=SANIC_ENV_PREFIX)
    config.project_name = project_name
    config.with_app = with_app

    if with_app:
        app_name = with_app
        app_package, app_tpl_dir, app_dst_dir = app_creation_params(
            app_name, proj_dst_dir)

        config.app_name = app_name
        config.app_package = app_package

    create_from_tpl(proj_tpl_dir, proj_dst_dir,
                    ignores=['*.pyc', ], **config)

    if with_app:
        create_app(ctx, app_name, app_package, app_tpl_dir, app_dst_dir)

    click.echo(f'New os-sanic project: {project_name}\n')
    click.echo('Use project template:')
    click.echo(f'    {proj_tpl_dir}')
    if with_app:
        click.echo('Use app template:')
        click.echo(f'    {app_tpl_dir}')
    click.echo('\n', nl=False)
    click.echo('Create project in:')
    click.echo(f'    {proj_dst_dir}')
    if with_app:
        click.echo('Create app in:')
        click.echo(f'    {app_dst_dir}')
    click.echo('\n', nl=False)
    click.echo('You can start your server with:')
    click.echo(f'    cd {proj_dst_dir}')
    click.echo('    python manager.py run')



import os
from os_sanic.utils import deep_update
from sanic.config import Config

SANIC_ENV_PREFIX = 'OS_SANIC_'


def create_sanic_config(defaults=None, load_env=True, keep_alive=True):
    default = SanicConfig(defaults, False, keep_alive=keep_alive)
    default_config_file = os.path.join(os.path.dirname(
        os.path.abspath(__file__)), 'default_config.py')
    default.from_pyfile(default_config_file)
    if not load_env:
        return default

    return SanicConfig(default, load_env, default.KEEP_ALIVE)


class SanicConfig(Config):

    def __init__(self, defaults=None, load_env=True, keep_alive=True):
        super(SanicConfig, self).__init__(defaults=defaults,
                                          load_env=load_env,
                                          keep_alive=keep_alive)
        self.LOGO = None



from collections import OrderedDict
from functools import partial
from inspect import isawaitable

from os_config import Config

from os_sanic.utils import load_class
from os_sanic.workflow import Workflowable


class Extension(Workflowable):

    def __init__(self, application, name, config):
        self.application = application
        self.name = name
        self.config = config
        self.logger = application.get_logger(name)

    @staticmethod
    def create(application, ext_cfg, user_config):
        ec = ext_cfg.extension_class
        config = Config.create()
        Config.update(config, ext_cfg)
        Config.update(config, user_config)
        name = Config.pop(config, 'name')
        Config.pop(config, 'extension_class')
        package = None
        if ec.startswith('.'):
            package = application.app_cfg.package
        cls = load_class(ec, Extension, package=package)
        return cls(application, name, config)


class ExtensionManager(Workflowable):
    def __init__(self, application):
        self.application = application
        self._extensions = OrderedDict()
        self.logger = application.get_logger(self.__class__.__name__)
        [setattr(self, method, partial(self.__call, method))
         for method in ('run', 'setup', 'cleanup')]

    def get_extension(self, name):
        return self._extensions[name]

    @property
    def extensions(self):
        return self._extensions.values()

    async def __call(self, method):
        iter = self._extensions.keys()
        if method == 'cleanup':
            iter = sorted(iter, reverse=True)

        for key in iter:
            ext = self._extensions[key]
            try:
                r = getattr(ext, method)()
                if isawaitable(r):
                    await r

            except Exception as e:
                self.logger.error(f'Extension error {key}.{method}, {e}')

    def load_extension(self, ext_cfg, user_config):

        try:
            name = ext_cfg.name
            if name in self._extensions:
                self.logger.warn(f'Extension already exists, {name}')
                return
            extension = Extension.create(
                self.application, ext_cfg, user_config)
            self._extensions[name] = extension
            self.logger.debug(f'Load extension, {name} {extension.__class__}')
        except Exception as e:
            self.logger.error(f'Load extension fail {e}, {ext_cfg}')

    @classmethod
    def create(cls, application):

        em = cls(application)

        user_configs = {}
        for cfg in Config.get(application.user_config, 'EXTENSIONS', []):
            name = Config.get(cfg, 'name')
            if name:
                user_configs[name] = cfg

        for ext_cfg in Config.get(application.core_config, 'EXTENSIONS', []):
            name = Config.get(ext_cfg, 'name')
            if name:
                em.load_extension(
                    ext_cfg, user_configs.get(name, Config.create()))
            else:
                logger = application.get_logger(cls.__name__)
                logger.warn(f'No name specified {ext_cfg}')

        return em



import os
import inspect
from collections import OrderedDict
from functools import partial
from importlib import import_module
from inspect import isawaitable

from os_config import Config

from os_sanic.extension import ExtensionManager
from os_sanic.log import getLogger
from os_sanic.view import ViewManager
from os_sanic.workflow import Workflowable


class Context(dict):
    def __setattr__(self, attr, value):
        self[attr] = value

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError as ke:
            raise AttributeError(f"No '{ke.args[0]}'")


class Application(Workflowable):

    def __init__(self, context):
        self.context = context

        self.logger = getLogger(f'App.{self.name}')

        self.extension_manager = ExtensionManager.create(self)

        self.view_manager = ViewManager.create(self)

        [setattr(self, method, partial(self.__call, method))
         for method in ('run', 'setup', 'cleanup')]

    @property
    def sanic(self):
        return self.application_manager.sanic

    def get_extension(self, extension_path):
        if '.' not in extension_path:
            extension_path = '.'.join((self.name, extension_path))

        return self.application_manager.get_extension(extension_path)

    def __getattr__(self, attr):
        try:
            if attr in self.context:
                return self.context[attr]
            return self[attr]
        except KeyError as ke:
            raise AttributeError("No '{}'".format(ke.args[0]))

    @property
    def name(self):
        return self.context.app_cfg.name

    def get_logger(self, tag):
        return getLogger(f'App.{self.name}.{tag}')

    async def __call(self, method):
        try:
            await getattr(self.extension_manager, method)()
        except Exception as e:
            self.logger.error(f'Method error {method}, {e}')

    @staticmethod
    def create(application_manager, app_cfg):

        app_module = import_module('.'.join((app_cfg.package, 'app')))
        runtime_path = os.path.dirname(app_module.__file__)

        core_config = Config.from_object(app_module)
        user_config = Config.create()
        if Config.get(app_cfg, 'config'):
            f = os.path.join(os.getcwd(), app_cfg.config)
            user_config = Config.from_pyfile(f)
            runtime_path = os.path.dirname(f)

        context = Context()
        context.application_manager = application_manager
        context.runtime_path = runtime_path
        context.app_cfg = app_cfg
        context.core_config = core_config
        context.user_config = user_config

        return Application(context)


class ApplicationManager(Workflowable):

    def __init__(self, sanic):
        self.sanic = sanic
        self.logger = getLogger(self.__class__.__name__)
        self._apps = OrderedDict()
        [setattr(self, method, partial(self.__call, method))
         for method in ('run', 'setup', 'cleanup')]

    async def __call(self, method):
        iter = self._apps.keys()
        if method == 'cleanup':
            iter = sorted(iter, reverse=True)

        for key in iter:
            app = self._apps[key]
            try:
                await getattr(app, method)()
            except Exception as e:
                self.logger.error(f'Application error {key}.{method}, {e}')

    def get_app(self, name):
        return self._apps[name]

    def get_extension(self, extension_path):
        app_name, extension_name = extension_path.split('.')
        app = self.get_app(app_name)
        return app.extension_manager.get_extension(extension_name)

    @property
    def apps(self):
        return self._apps.values()

    def _load_app(self, app_cfg):
        if isinstance(app_cfg, str):
            app_cfg = Config.create(package=app_cfg)
        if not Config.get(app_cfg, 'package'):
            self.logger.warn(f'Skip, no package {app_cfg}')
            return

        app_name = app_cfg.package.split('.')[-1]
        app_name = Config.get(app_cfg, 'name', app_name)
        if not app_name:
            self.logger.warn(f'Skip, no valid app name {app_cfg}')
            return
        app_cfg.name = app_name
        if app_name in self._apps:
            self.logger.warn(f'Skip, app already exists, {app_name}')
            return

        self.logger.debug(
            f'Load app, {app_name} <package \'{app_cfg.package}>\'')
        app = Application.create(self, app_cfg)
        self._apps[app_name] = app

    def load_app(self, app_cfg):
        try:
            self._load_app(app_cfg)
        except Exception as e:
            self.logger.error(f'Load app fail, {e}, {app_cfg}')

    @classmethod
    def create(cls, sanic):

        assert not hasattr(
            sanic, 'application'), 'sanic instance already has \'application\' attribute'

        am = cls(sanic)
        for app_cfg in Config.create(
                apps=sanic.config.get('INSTALLED_APPS', [])).apps:
            am.load_app(app_cfg)
        return am



import os
import shlex
import subprocess
import sys

import pytest
from sanic import Sanic

from os_sanic.manage import main


def process(runable_file=None, cmdline='', env=None, **kwargs):
    if env is None:
        env = os.environ.copy()
    if env.get('COVERAGE', None) is not None:
        if 'COVERAGE_PROCESS_START' not in env:
            env['COVERAGE_PROCESS_START'] = os.path.abspath('.coveragerc')

    f = runable_file
    if not f:
        f = os.path.abspath(__file__)
    cmd = 'python -u {} {}'.format(f, cmdline)
    proc = subprocess.Popen(shlex.split(cmd),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.PIPE,
                            cwd=os.getcwd(),
                            env=env,
                            **kwargs)
    return proc


def call(runable_file=None, cmdline='', env=None, **kwargs):
    proc = process(runable_file, cmdline, env, **kwargs)
    stdout, stderr = proc.communicate()
    return stdout, stderr


if __name__ == "__main__":
    sys.path.insert(0, os.getcwd())
    if os.getenv('COVERAGE_PROCESS_START'):
        import coverage
        coverage.process_startup()
    main()



import os

import click

from os_sanic.commands import valid_log_level
from os_sanic.config import SANIC_ENV_PREFIX, create_sanic_config
from os_sanic.server import Server


def create_server(app, **kwargs):
    config_file = kwargs.get('config_file', None)
    if config_file and not isinstance(config_file, str):
        config_file = config_file.name

    config = create_sanic_config(load_env=SANIC_ENV_PREFIX)
    if config_file:
        config_file = os.path.abspath(config_file)
        kwargs['config_file'] = config_file
        config.from_pyfile(config_file)
    config.update(dict([(k.upper(), v) for k, v in kwargs.items()])),

    server = Server.create(
        app,
        config=config,
        env_prefix=SANIC_ENV_PREFIX,
    )
    return server


default_config = create_sanic_config(load_env=SANIC_ENV_PREFIX)


@click.command()
@click.option('--access-log', is_flag=True, help='Enable access log.')
@click.option('--debug', is_flag=True, help='Enable debug mode.')
@click.option('-c', '--config-file',
              default='config.py', show_default=True,
              type=click.File(mode='r'), help='Config file.')
@click.option('-h', '--host',
              default=default_config.get('HOST'),
              show_default=True, help='Bind address.')
@click.option('-p', '--port',
              default=default_config.get('PORT'),
              show_default=True, help='Listen port.')
@click.option('-l', '--log-level', metavar='LOG_LEVEL',
              default=default_config.get('LOG_LEVEL'), show_default=True,
              callback=valid_log_level,
              help='Log level.'
              )
@click.pass_context
def cli(ctx, **kwargs):
    '''Run server.'''

    ctx.ensure_object(dict)

    server = create_server(ctx.obj['app'], **kwargs)
    server.run()



EXTENSIONS = [
    {
        'name': 'case02',
        'extension_class': '.extension.Case02Extension'
    }
]

VIEWS = [
    {
        'uri': '/',
        'view_class': '.view.Case02View'
    },
]

STATICS = [
    {
        'uri': '/static',
        'file_or_directory': './static'
    }
]
